#include <Ports.h>
#include <RF12.h>
#include <avr/eeprom.h>

//const byte totalDevices = 3;
const byte totalDevices = 2;

//Port AIn(1);
//Port DOut(2);
//Port DIn(3);

Port DIn(1);
Port DOut(2);


boolean sendInit; // has init been send and received by HomeSeer
boolean sendDev;  // have alle dev values been send and received by HS
boolean sendDevValues[totalDevices]; // same as above, only for each device value
boolean needToSend;

MilliTimer sendTimer;


/*************************************************************************
 *                                                                       *
 *                       HomeSeer Code starts here                       *
 *                                                                       *
 *************************************************************************/

/***************************************************************
 * Save config section                                         *
 * Derived from RF12demo of rf12library by Jean-Claude Wippler *
 * See http://jeelabs.org/view/libraries/RF12/                 *
 ***************************************************************/

typedef struct {
  byte nodeId;
  byte group;
  char msg[RF12_EEPROM_SIZE-4];
  word crc;
} 
RF12Config;

static RF12Config config;

static void addCh (char* msg, char c) {
  byte n = strlen(msg);
  msg[n] = c;
}

static void addInt (char* msg, word v) {
  if (v >= 10)
    addInt(msg, v / 10);
  addCh(msg, '0' + v % 10);
}

#define COLLECT 0x20 // collect mode, i.e. pass incoming without sending acks
#include <util/crc16.h>

static void saveConfig () {
  // set up a nice config string to be shown on startup
  memset(config.msg, 0, sizeof config.msg);
  strcpy(config.msg, " ");

  byte id = config.nodeId & 0x1F;
  addCh(config.msg, '@' + id);
  strcat(config.msg, " i");
  addInt(config.msg, id);
  if (config.nodeId & COLLECT)
    addCh(config.msg, '*');

  strcat(config.msg, " g");
  addInt(config.msg, config.group);

  strcat(config.msg, " @ ");
  static word bands[4] = { 
    315, 433, 868, 915       };
  word band = config.nodeId >> 6;
  addInt(config.msg, bands[band]);
  strcat(config.msg, " MHz ");

  config.crc = ~0;
  for (byte i = 0; i < sizeof config - 2; ++i)
    config.crc = _crc16_update(config.crc, ((byte*) &config)[i]);

  // save to EEPROM
  for (byte i = 0; i < sizeof config; ++i) {
    byte b = ((byte*) &config)[i];
    eeprom_write_byte(RF12_EEPROM_ADDR + i, b);
  }

  if (!rf12_config())
    Serial.println("config save failed");
  else
    Serial.println("Config saved");
}

/**********************
 * End config section *
 **********************/

/***********************************************************
 * Class for a device                                      *
 * A JeeNode can have up to eight devices. Each device     *
 * has a status and a few other properties (input, output) *
 ***********************************************************/


/* Device: 
 * String of devicename
 * type of device (input, output, etc)
 * current status
 * number of status
 * statusChanged (after update, changed since last time?)
 * overRuled (should status altered by decisionRules?)
 * notificationType: push or noPush. Push means direct update to JeeLink/HomeSeer after a change
 * storagearray for statuses
 *
 * Function: setStatus(byte i). Sets status to byte i and compares current status with previous status
 */
class Device {
  // Devicestatus: String of the status and associated value
  struct DevStatus {
    char statusName[5]; // maxlength: 4
    byte value; 
  };

public:
  static const byte DigIn = 0;
  static const byte DigOut = 1;
  static const byte AnaIn = 2;
  static const byte AnaOut = 3;

  static const boolean push = true;
  static const boolean noPush = false;

  char name[5]; //maxlength: 4
  byte devType;
  int currentStatus;
  byte amountStatus; //max 8
  DevStatus statusArray[8];
  Port* outputPort;
  Port* inputPort;

//  void setStatus(int i) {
//    currentStatus = i; 
//  }

  void setStatus(int i) {
    
    currentStatus = i; 
    
    switch (devType) {
      
      case DigOut:
        outputPort->digiWrite((byte)currentStatus);
        break;

      case AnaOut:
        outputPort->anaWrite(currentStatus);
        break;
    }
  }

  void getStatus() {

    int newStatus = currentStatus;
    
    switch (devType) {
      
      case DigIn:
        newStatus = inputPort->digiRead();
        break;

      case AnaIn:
        newStatus = inputPort->anaRead();
        break;
    }
    
    if (newStatus != currentStatus) needToSend = true;
    
    currentStatus = newStatus; 
    
  }
  
};

Device devices[totalDevices]; // Storage of eight devices

/**********************
 * End device section *
 **********************/

/*
 * Array copy is an extra function to copy strings intro an string variable.
 * strcpy() is not sufficient, as remaining space should be filled up with spaces
 */
void arrayCopy(char target[], char source[]) {
  int i;
  for(i=0; i<4; i++) {
    target[i] = ' ';
    if(source[i] == '\0')
      break;
    else
      target[i] = source[i];
  }
  // fill up extra spaces 
  for(;i<4;i++) {
    target[i] = ' '; 
  }
  target[4] = '\0';
}


/******************
 * Config section *
 ******************/
 
void setup() {
  Serial.begin(57600);
  Serial.println("\n[JeeNode_for_Homeseer_LED]");

  //===============================//
  // config devices

  //device 1: Type Digital In
  arrayCopy(devices[0].name, "DIn");
  devices[0].devType = Device::DigIn;
  devices[0].currentStatus = 0;
  devices[0].amountStatus = 2;
  arrayCopy(devices[0].statusArray[0].statusName, "Off");    
  devices[0].statusArray[0].value = 0;  
  arrayCopy(devices[0].statusArray[1].statusName, "On");    
  devices[0].statusArray[1].value = 1;
  devices[0].inputPort = &DIn;

   //device 2: Digital Out
  arrayCopy(devices[1].name, "DOut");
  devices[1].devType = Device::DigOut;
  devices[1].currentStatus = 0;
  devices[1].amountStatus = 2;
  arrayCopy(devices[1].statusArray[0].statusName, "Off");    
  devices[1].statusArray[0].value = 0;  
  arrayCopy(devices[1].statusArray[1].statusName, "On");    
  devices[1].statusArray[1].value = 1;
  devices[1].outputPort = &DOut;
  



  // config of radio & transmission booleans
  if (rf12_config()) {
    config.nodeId = eeprom_read_byte(RF12_EEPROM_ADDR);
    config.group = eeprom_read_byte(RF12_EEPROM_ADDR + 1);
  } 
  else {
    config.nodeId = 0x41; // node A1 @ 433 MHz
    config.group = 0xD4;
    saveConfig();
  }
  rf12_encrypt(RF12_EEPROM_EKEY);

  needToSend = true;
  sendInit = false;
  sendDev = false;
  for(int i=0; i<totalDevices; i++) {
    sendDevValues[i] = false; 
  }

  // other config, application specific
//  AIn.mode(INPUT);
//  DIn.mode(INPUT);
//  DIn.mode2(INPUT);
//  DOut.mode(OUTPUT);
    DIn.mode2(INPUT);
    DOut.mode(OUTPUT);

}


/**********************
 * End config section *
 **********************/

void loop() {

  // try to send Init
  if(!sendInit) {
    sendInitMessage();
  }

  // try to send the device values
  if(!sendDev && sendInit) {
    boolean sendAll = true;
    for(byte i=0; i<totalDevices; i++) {
      if(!sendDevValues[i]) {
        sendDevValue(i);
      }
      sendAll &= sendDevValues[i];
    }
    if(sendAll) {
      sendDev = true;
    }   
  }

  // Check for new incoming messages
  checkForInput();
  
  // read input ports
  updateInputDevices();
  

  if (needToSend && rf12_canSend() && sendInit && sendDev) {
    sendStatuses();
  }
}

/*****************
 * Input section *
 *****************/
 
// Check for new input on the radio and serial line
void checkForInput() {

  // Read data from Serial-in
  if (Serial.available())
    handleInput(Serial.read());

  // check for new data on radio
  if (rf12_recvDone() == 1) {
    Serial.println("received data");
    if(rf12_crc == 0) {
      Serial.print(" ");
      Serial.println("ok data");
      processData(); // only correctly radio received data
    }
  }
}

/* 
 * Derived from RF12demo of rf12library by Jean-Claude Wippler
 * See http://jeelabs.org/view/libraries/RF12/
 */

static byte value, stack[RF12_MAXDATA], top;
static void handleInput (char c) {
  if ('0' <= c && c <= '9')
    value = 10 * value + c - '0';
  else if (c == ',') {
    if (top < sizeof stack)
      stack[top++] = value;
    value = 0;
  } 
  else if ('a' <= c && c <='z') {

    switch (c) {
    default:
      Serial.println("Wrong input");
      break;
    case 'c': // Config encryption key
      stack[top++] = value;
      configEncrKey();
      break;
    case 'n': // save node id : nodeId,GroupId, Frequency
      config.nodeId = (config.nodeId & 0xE0) + (stack[0] & 0x1F);
      config.group = stack[1];
      value = value == 8 ? RF12_868MHZ :
      value == 9 ? RF12_915MHZ : RF12_433MHZ;
      config.nodeId = (RF12_868MHZ << 6) + (config.nodeId & 0x3F);
      saveConfig();
      break;
    }
    value = top = 0;
    memset(stack, 0, sizeof stack);
  }
}

// save encryption key
void configEncrKey() {
  eeprom_write_block(stack, RF12_EEPROM_EKEY, 16);
  rf12_encrypt(RF12_EEPROM_EKEY);
}

/*********************
 * End input section *
 *********************/

/*******************
 * Message section *
 *******************/

#define initMessage 0
#define devValues 1
#define statusUpdate 2
#define decisionRule 3
#define overRule 4

#define receivedInit 50
#define receivedDevValues 51

#define requestInit 100
#define requestDevValues 101
#define updateDecRule 102
#define requestDecRules 103

/*********************
 * Outgoing Messages *
 *********************/

/*
 * Send init message: what for devices on JeeNode?
 *
 * type of messsage: initMessage
 * Number of devices
 * Foreach device: name + type
 */
void sendInitMessage() {

  byte toSend[2+totalDevices*5];
  toSend[0] = initMessage;
  toSend[1] = totalDevices;
  byte k = 2;
  for(int i=0; i<totalDevices; i++) {
    toSend[k++] = devices[i].name[0];
    toSend[k++] = devices[i].name[1];
    toSend[k++] = devices[i].name[2];
    toSend[k++] = devices[i].name[3];  
    toSend[k++] = devices[i].devType;
  }

  // send message until received by HomeSeer.
  // Keep checking for input and updating the system
  boolean maySend = true;
  while (!sendInit) {
    checkForInput();

    if(sendTimer.poll(5000)) {
      maySend = true; 
    }

    if(maySend && rf12_canSend()) {
      rf12_sendStart(0, &toSend, sizeof toSend);
      maySend = false;
      Serial.println(" Init zenden");
    }
  }
}

/*
 * Send deviceValues: What message comes with each value
 *
 * type of message: devValues
 * DeviceNumber
 * devType
 * currentStatus
 * amountStatus
 * foreach status: Name + value
 */

void sendDevValue(byte i) {
  Device dev = devices[i];
  byte toSend[6+dev.amountStatus*6];
  toSend[0] = devValues;
  toSend[1] = i;
  toSend[2] = dev.devType;
  toSend[3] = dev.currentStatus >> 8 & 0xff; 
  toSend[4] = dev.currentStatus & 0xff;
  toSend[5] = dev.amountStatus;

  byte k=6;
  for(byte j=0; j<dev.amountStatus; j++) {

    toSend[k++] = dev.statusArray[j].statusName[0];
    toSend[k++] = dev.statusArray[j].statusName[1];
    toSend[k++] = dev.statusArray[j].statusName[2];
    toSend[k++] = dev.statusArray[j].statusName[3];
    toSend[k++] = dev.statusArray[j].value >> 8; 
    toSend[k++] = dev.statusArray[j].value;
  }

  // send message until received by HomeSeer.
  // Keep checking for input and updating the system
  boolean maySend = true;
  while (!sendDevValues[i]) {
    checkForInput();

    if(sendTimer.poll(5000)) {
      maySend = true; 
    }

    if(maySend && rf12_canSend()) {
      rf12_sendStart(0, &toSend, sizeof toSend);
      maySend = false;
      Serial.print("DevValues ");
      Serial.print((int) i);
      Serial.println(" zenden");
    }
  }
}

/*
 * Send a status message
 * Type of message: statusUpdate
 * foreach device the status as int (= 2 bytes)
 */
void sendStatuses() {
  byte toSend[1+totalDevices*2];
  toSend[0] = statusUpdate;
  byte k=1;
  for(int i=0; i<totalDevices; i++) {
    toSend[k++] = devices[i].currentStatus >> 8; 
    toSend[k++] = devices[i].currentStatus;
  }
  rf12_sendStart(0, &toSend, sizeof toSend);
  Serial.println("zenden"); 

  needToSend = false;
}

void updateInputDevices() {
  for(byte i=0; i<totalDevices; i++) {
    devices[i].getStatus();
  } 
}


/*********************
 * Incoming messages *
 *********************/


/*
 * ProcessData: process incoming data
 */
void processData() {
  switch(rf12_data[0]) {
  case requestInit:
    sendInit = false; // message consists of 1 byte messageType.
    break;
  case requestDevValues:
    sendDevValue(rf12_data[1]); // message consists of 1 byte messageType, 1 byte DeviceValue
    break;
  case receivedInit:
    sendInit = true; // Init has been received by HomeSeer
    break;
  case receivedDevValues:
    sendDevValues[rf12_data[1]] = true; // Device Value i has been received by HomeSeer
    break;
  case overRule:
    overRuleMessage();
  default:
    break;
  } 
}



/*
 * Received a message to overrule or stop overruling a device
 *
 * Type of message (handled by processData())
 * number of the device
 * parameter to overrule:
 *  0: always off
 *  1: always on
 *  2: reset and listen to decisionRules
 */

void overRuleMessage() { 
  devices[rf12_data[1]].setStatus(rf12_data[2]);
  // yeah, 2-way! :-)
  needToSend = true;  
}

